from requests_futures.sessions import FuturesSession
from concurrent.futures import CancelledError, TimeoutError
import logging
import json
import time

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)

class publisherPost():
    """Publish message using POST.
    
    Parameters
    ----------
    url: str
        URL to post messages to
    api_key: str
        API key
    """
    def __init__(self, url, auth_url, api_key):
        self.url = url
        self.session = FuturesSession()
        self.auth_url = auth_url
        self.api_key = api_key
        self.refreshing_token = None
        self.access_token = self._get_access_token()
        self.timeout = 60
        self.max_retries = 3
    
    def _get_access_token(self):
        print('Getting access token')
        logger.warn('Getting access token')
        future = FuturesSession().post(
            self.auth_url, json={'apiKey': self.api_key}
        )
        self.refreshing_token = future
        
        try:
            response = future.result()
            access_token = json.loads(response.content)['access_token']
        except (CancelledError, TimeoutError, Exception) as e:
            access_token = None
            logger.warn('Could not get access token - error = {}'.format(e))
        return access_token
    
    def emit(self, message, callback, nb_retries = 0):
        """Publish message
        
        Parameters
        ----------
        message: dict
            Message to push to Stakion.
        callback: func
            The callback function is called when the message is acknowledges.
        """
        def response_hook(resp, *args, **kwargs):
            logger.info('status_code = {} - nb_retries = {}'.format(resp.status_code, nb_retries))
            if resp.status_code == 200:
                callback(message['__uuid'])
            elif (resp.status_code == 401) & (nb_retries < self.max_retries):
                error = json.loads(resp.content)['error']
                if error == 'Expired access token':
                    
                    if self.refreshing_token.done():
                        self.access_token = self._get_access_token()
                        self.emit(message, callback, nb_retries + 1)
                    else:
                        # Wait for token refresh
                        logger.info('Waiting for refresh token')
                        while self.refreshing_token.done() == False:
                            time.sleep(0.1)
                        
                        # Token is either refreshed or hasn't updated
                        self.emit(message, callback, nb_retries + 1)
                
                else:
                    logger.warning('Could not send message')
            else:
                logger.warning('Could not send message - uid = ${}'.format(message['__uuid']))
        
        self.session.headers['authorization'] = self.access_token
        future = self.session.post(self.url, json={'log': message}, hooks={
            'response': response_hook
        })


# -*- coding: utf-8 -*-
import time
import traceback
import json
import uuid
import warnings
import socketio

from stakion import publishers

class StakionSender(object):
    """Stakion logger.
    
    Parameters
    ----------
    api_key: str
        Api Key as generated in the Stakion dashboard.
    project_id: str
        Project ID as generated in the Stakion dashboard.
    host: str, optional
        Target host. Does not need to be supplied in the majority of cases.
    bufmax: int, optional
        Maximum number of bytes in the message queue before the
        buffer overflow handler is called. This can happen when messages don't
        get acknowledged by Stakion. Default is 1.024MB.
    buffer_overflow_handler: func, optional
        Buffer overflow handler is called when bufmax is exceeded. A list of
        dictionaries is passed.
    method: str, optional
        The method to send the messages. Currently the only supported values
        is POST.
    
    """
    def __init__(self,
                 api_key,
                 project_id,
                 host=None,
                 bufmax=1 * 1024 * 1024,
                 buffer_overflow_handler=None,
                 method='POST'):
        self.api_key = api_key
        self.project_id = project_id
        self.bufmax = bufmax
        self.buffer_overflow_handler = buffer_overflow_handler
        
        self.pendings = {}
        
        # Unsupported method
        if method not in ['POST']:
            raise ValueError("Method is not supported - valid values are: [POST]")
        
        if host is None:
            if method == 'POST':
                auth_url = 'https://mamooh-223220.appspot.com/auth'
                url = 'https://mamooh-223220.appspot.com/log/' + project_id
        else:
            if method == 'POST':
                if host[-1] == '/':
                    url = host + 'log/' + project_id
                    auth_url = host + 'auth'
                else:
                    url = url + '/log/' + project_id
                    auth_url = host + '/auth'
        
        # Initialise the publisher
        if method == 'POST':
            self.publisher = publishers.publisherPost(
                url=url,
                auth_url=auth_url,
                api_key=api_key
            )
    
    def publish(self, correlation_id, event_type, model, data=None):
        """Push log message for analysis.
        
        Parameters
        ----------
        correlation_id: str
            The correlation ID identifies a unique prediction and will be
            used to link mupltiple log messages to a single prediction. This
            is especially useful to link input features to output features.
        event_type: str
            Log type, supported values are: prediction_requested,
            input_features, output_features, crash_logger, error.
        model: str
            Model name
        data: dict, optional
            Log message.
        """
        timestamp = int(round(time.time() * 1000))
        
        try:
            msg = self._make_message(
                correlation_id=correlation_id,
                event_type=event_type,
                model=model,
                timestamp=timestamp,
                data=data)
        except Exception as e:
            data = {
                "level": "CRITICAL",
                "message": "Can't output to log",
                "traceback": traceback.format_exc()
            }
            msg = self._make_message(
                correlation_id=correlation_id,
                event_type="crash_logger",
                model=model,
                timestamp=timestamp,
                data=data)
        
        return self._send(msg)
    
    def _make_message(self, correlation_id, event_type, model, timestamp, data):
        """Create and validate message from log.
        
        Parameters
        ----------
        correlation_id: str
        event_type: str
        model: str
        timestamp: int
        data: dict
        """
        # Create message
        message = {'projectId': self.project_id,
                   'correlation_id': correlation_id,
                   'event_type': event_type,
                   'model': model,
                   'timestamp': timestamp,
                   'data': data}
        
        # Validate message
        self._validate_message(message)
        
        return message
    
    def _validate_message(self, message):
        """Validate a message
        
        This function currently validates the event_type only.
        
        Parameters
        ----------
        message: dict
            Message as generated by _make_message
        """
        # Event types
        supported_event_types = ['prediction_requested', 'input_features',
                                 'output_features', 'crash_logger', 'error']
        
        event_type = message['event_type']
        if event_type not in supported_event_types:
            raise warnings.warn(
                'Event type not supported: The event type ' +
                event_type + ' was passed ' +
                'but only the following event types are '+
                'supported ' + ', '.join(supported_event_types) +
                '. As such no aggregation will be performed.'
            )
    
    def _remove_pending(self, id):
        """Remove message from pending list following acknowledgement.
        
        Parameters
        ----------
        id: str
            Message id
        """
        try:
            self.pendings.pop(id)
        except KeyError:
            print('Issue removing pending message')
    
    def _send(self, message):
        """Send formated log message.
        
        Parameters
        ----------
        message: dict
        """
        # clear buffer if it exceeds max buffer size
        if len(self.pendings.keys()) > 0:
            bufsize = sum(len(json.dumps(s).encode('utf-8')) for s in self.pendings.values())
            if (bufsize > self.bufmax):
                messages_in_buffer = self.pendings.values()
                self._call_buffer_overflow_handler(messages_in_buffer)
                self.pendings = {}
        
        # Create unique message id
        message_id = uuid.uuid4().hex
        message['__uuid'] = message_id
        
        self.pendings[message_id] = message
        self.publisher.emit(message=message, callback=self._remove_pending)
    
    def _call_buffer_overflow_handler(self, pending_events):
        """Call buffer overflow handler."""
        try:
            if self.buffer_overflow_handler:
                self.buffer_overflow_handler(pending_events)
        except Exception as e:
            # User should care any exception in handler
            pass
    


import os
import sys

from setuptools import setup
from setuptools.command.install import install

VERSION = "0.0.5"

def readme():
    """print long description"""
    with open('README.rst') as f:
        return f.read()

class VerifyVersionCommand(install):
    """Custom command to verify that the git tag matches our version"""
    description = 'verify that the git tag matches our version'

    def run(self):
        tag = os.getenv('CIRCLE_TAG')

        if tag != VERSION:
            info = "Git tag: {0} does not match the version of this app: {1}".format(
                tag, VERSION
            )
            sys.exit(info)
 
#from distutils.core import setup
#import setuptools

setup(
    name='stakion-logger',
    version=VERSION,
    long_description=readme(),
    url='https://stakion.io/',
    author='Jacques Verre',
    author_email='jacques.verre@stackion.io',
    license='LICENSE.txt',
    description='Logger for stakion.',
    packages=['stakion'],
    install_requires=[
          'requests_futures',
    ],
    python_requires='>=2.7,!=3.0.*,!=3.1.*,!=3.2.*,!=3.3.*',
    cmdclass={
        'verify': VerifyVersionCommand,
    }
)